<?php

declare(strict_types=1);

namespace Drupal\data_pipelines_opensearch;

/**
 * A class that represents a document object.
 *
 * Used as a value object for document requests.
 *
 * @package Drupal\data_pipelines_opensearch\Request\Object
 */
class Document {

  /**
   * The constructor for Document.
   *
   * @param string $id
   *   The document ID.
   * @param Index $index
   *   The document index.
   * @param array $data
   *   The document data.
   * @param int|null $delta
   *   The document delta (for tracking purposes).
   */
  public function __construct(protected string $id, protected Index $index, protected array $data, protected ?int $delta = NULL) {}

  /**
   * A method to get the document ID.
   *
   * @return string
   *   The document ID.
   */
  public function getId(): string {
    return $this->id;
  }

  /**
   * A method to set the document ID.
   *
   * @param string $id
   *   The document ID.
   *
   * @return $this
   *   The document.
   */
  public function setId(string $id): self {
    $this->id = $id;
    return $this;
  }

  /**
   * A method to get the document index.
   *
   * @return \Drupal\data_pipelines_opensearch\Index
   *   The document index.
   */
  public function getIndex(): Index {
    return $this->index;
  }

  /**
   * A method to set the document index.
   *
   * @param \Drupal\data_pipelines_opensearch\Index $index
   *   The document index.
   *
   * @return $this
   *   The document.
   */
  public function setIndex(Index $index): self {
    $this->index = $index;
    return $this;
  }

  /**
   * A method to get the document data.
   *
   * @return array
   *   The document data.
   */
  public function getData(): array {
    return $this->data + [
      '@delta' => $this->delta,
    ];
  }

  /**
   * A method to set the document data.
   *
   * @param array $data
   *   The document data.
   *
   * @return $this
   *   The document.
   */
  public function setData(array $data): self {
    $this->data = $data;
    return $this;
  }

  /**
   * A method to get the document delta.
   *
   * @return int|null
   *   The document delta.
   */
  public function getDelta(): ?int {
    return $this->delta;
  }

}
