<?php

declare(strict_types=1);

namespace Drupal\data_pipelines_opensearch\Plugin\DatasetDestination;

use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\data_pipelines\Attribute\DatasetDestination;
use Drupal\data_pipelines\DatasetPipelinePluginManager;
use Drupal\data_pipelines\Destination\DatasetDestinationPluginBase;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\data_pipelines\Entity\DestinationInterface;
use Drupal\data_pipelines_opensearch\ClientFactoryInterface;
use Drupal\data_pipelines_opensearch\Document;
use Drupal\data_pipelines_opensearch\DocumentsRequest;
use Drupal\data_pipelines_opensearch\Index;
use Drupal\data_pipelines_opensearch\IndexRequest;
use OpenSearch\Client;
use OpenSearch\Exception\BadRequestHttpException;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A class that represents the OpenSearch destination plugin.
 *
 * @package Drupal\data_pipelines_opensearch\Plugin\DatasetDestination
 */
#[DatasetDestination(
  'opensearch',
  new TranslatableMarkup('OpenSearch'),
)]
class OpenSearch extends DatasetDestinationPluginBase implements ContainerFactoryPluginInterface {

  /**
   * The constructor for OpenSearch.
   *
   * @param array $configuration
   *   The plugin configuration.
   * @param string $plugin_id
   *   The plugin ID.
   * @param mixed $plugin_definition
   *   The plugin definition.
   * @param \Drupal\data_pipelines\DatasetPipelinePluginManager $datasetPipelinePluginManager
   *   The dataset pipeline plugin manager.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $loggerChannel
   *   The logger channel.
   * @param \Drupal\data_pipelines_opensearch\ClientFactoryInterface $clientFactory
   *   The client factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, protected DatasetPipelinePluginManager $datasetPipelinePluginManager, protected LoggerChannelInterface $loggerChannel, protected ClientFactoryInterface $clientFactory, protected EntityTypeManagerInterface $entityTypeManager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.data_pipelines_pipeline'),
      $container->get('logger.channel.data_pipelines_opensearch'),
      $container->get('data_pipelines_opensearch.client_factory'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function viewSettings(): array {
    $client =
      $this->clientFactory->create(
        $this->getUrl(),
        $this->getUsername(),
        $this->getPassword()
      );
    return [
      'data' => [
        '#markup' => new TranslatableMarkup("%url %ok<br/>Prefix: %prefix", [
          '%url' => $this->getUrl(),
          '%ok' => $client && $client->ping() ? "✅ OK" : "❌ Failed",
          '%prefix' => $this->getPrefix(),
        ]),
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state): array {
    $form['url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('URL'),
      '#description' => $this->t('URL of the OpenSearch cluster.'),
      '#default_value' => $this->configuration['url'],
      '#required' => TRUE,
    ];
    $form['username'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Username'),
      '#description' => $this->t('Username to authenticate with.'),
      '#default_value' => $this->configuration['username'],
    ];
    $form['password'] = [
      '#type' => 'password',
      '#title' => $this->t('Password'),
      '#description' => $this->t('Password to authenticate with.'),
      '#default_value' => $this->configuration['password'],
    ];
    $form['prefix'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Index prefix'),
      '#description' => $this->t('The index prefix to use.'),
      '#default_value' => $this->configuration['prefix'],
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state): void {
    if (!UrlHelper::isValid($form_state->getValue('url'), TRUE)) {
      $form_state->setErrorByName('url', $this->t("URL must be a valid format."));
    }
  }

  /**
   * A method to get the configured URL.
   *
   * @return string
   *   The configured URL.
   */
  public function getUrl(): string {
    return $this->configuration['url'];
  }

  /**
   * A method to get the configured username.
   *
   * @return string
   *   The configured username.
   */
  public function getUsername(): string {
    return $this->configuration['username'];
  }

  /**
   * A method to get the configured password.
   *
   * @return string
   *   The configured password.
   */
  public function getPassword(): string {
    return $this->configuration['password'];
  }

  /**
   * A method to get the configured prefix.
   *
   * @return string
   *   The configured prefix.
   */
  public function getPrefix(): string {
    return $this->configuration['prefix'];
  }

  /**
   * A method to get the index name.
   *
   * This is composed of the prefix and dataset machine name.
   *
   * @param \Drupal\data_pipelines\Entity\DatasetInterface $dataset
   *   The dataset.
   *
   * @return string
   *   The index name.
   */
  public function getIndexName(DatasetInterface $dataset): string {
    return $this->getPrefix() . $dataset->getMachineName();
  }

  /**
   * A method to get create an instance of the OpenSearch client.
   *
   * @return \OpenSearch\Client|null
   *   The OpenSearch client.
   */
  protected function getClient(): ?Client {
    return $this->clientFactory->create(
        $this->getUrl(),
        $this->getUsername(),
        $this->getPassword()
      );
  }

  /**
   * {@inheritdoc}
   */
  public function deleteDataSet(DatasetInterface $dataset, DestinationInterface $destination): bool {
    if (!$client = $this->getClient()) {
      return FALSE;
    }
    $index_request = new IndexRequest($client);
    $index = new Index($this->getIndexName($dataset));
    if (!$index_request->exists($index)) {
      return TRUE;
    }
    $index_request->delete($index);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function beginProcessing(DatasetInterface $dataset): bool {
    parent::beginProcessing($dataset);
    if (!$client = $this->getClient()) {
      return FALSE;
    }
    $settings = $dataset->getPipeline()->getDestinationSettings($this->getPluginId());
    $index_request = new IndexRequest($client);
    $index = new Index($this->getIndexName($dataset), $settings);
    if ($index_request->exists($index)) {
      return TRUE;
    }
    try {
      $index_request->create($index);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->loggerChannel->error($e->getMessage());
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function processChunk(DatasetInterface $dataset, array $chunk): bool {
    parent::processChunk($dataset, $chunk);
    if (!$client = $this->getClient()) {
      return FALSE;
    }
    $documents_request = new DocumentsRequest($client);
    $index = new Index($this->getIndexName($dataset));
    $documents = [];
    foreach ($chunk as $delta => $item) {
      $id = "{$dataset->getMachineName()}:{$delta}";
      $documents[] = new Document($id, $index, $item->getArrayCopy(), $delta);
    }
    try {
      $documents_request->create($index, $documents);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->loggerChannel->error($e->getMessage());
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function processCleanup(DatasetInterface $dataset, array $invalidDeltas): bool {
    parent::processCleanup($dataset, $invalidDeltas);
    if (!$client = $this->getClient()) {
      return FALSE;
    }
    $documents_request = new DocumentsRequest($client);
    $index = new Index($this->getIndexName($dataset));
    $documents = [];
    foreach ($invalidDeltas as $delta) {
      $id = "{$dataset->getMachineName()}:{$delta}";
      $documents[] = new Document($id, $index, [], $delta);
    }
    try {
      $documents_request->delete($index, $documents);
      return TRUE;
    }
    catch (\Exception $e) {
      $this->loggerChannel->error($e->getMessage());
    }
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function getLastDelta(DatasetInterface $dataset): int {
    if (!$client = $this->getClient()) {
      return 0;
    }
    $index_request = new IndexRequest($client);
    $documents_request = new DocumentsRequest($client);
    $index = new Index($this->getIndexName($dataset));
    if ($index_request->exists($index)) {
      try {
        if ($document = $documents_request->last($index)) {
          return $document->getDelta();
        }
      }
      catch (BadRequestHttpException $e) {
        $this->loggerChannel->warning($e->getMessage());
      }
    }
    return 0;
  }

}
