<?php

declare(strict_types=1);

namespace Drupal\Tests\data_pipelines_opensearch\Kernel;

use Drupal\data_pipelines\Destination\ProcessingOperationEnum;
use Drupal\data_pipelines\Entity\DatasetInterface;
use Drupal\data_pipelines\Entity\DestinationInterface;
use Drupal\data_pipelines\Form\DatasetBatchContext;
use Drupal\data_pipelines_opensearch\Index;
use Drupal\data_pipelines_opensearch\IndexRequest;
use Drupal\Tests\data_pipelines\Traits\DatasetTestProcessingTrait;

/**
 * A class to test operations against the destination.
 *
 * @package Drupal\Tests\data_pipelines_opensearch\Kernel
 *
 * @group data_pipelines_opensearch
 */
class OpenSearchDestinationTest extends OpenSearchTestBase {

  use DatasetTestProcessingTrait;

  /**
   * The dataset.
   *
   * @var \Drupal\data_pipelines\Entity\DatasetInterface|null
   */
  private ?DatasetInterface $dataset = NULL;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();
    $this->dataset = $this->createTestOpenSearchDataset();
  }

  /**
   * {@inheritdoc}
   */
  protected function tearDown(): void {
    $this?->dataset->delete();
    parent::tearDown();
  }

  /**
   * A method to get the dataset.
   *
   * @return \Drupal\data_pipelines\Entity\DatasetInterface
   *   The dataset.
   */
  private function getDataset(): DatasetInterface {
    return $this->dataset;
  }

  /**
   * A method to get the dataset destination.
   *
   * @return \Drupal\data_pipelines\Entity\DestinationInterface
   *   The dataset destination.
   */
  private function getDestination(): DestinationInterface {
    $destinations = $this->getDataset()->getDestinations();
    return reset($destinations);
  }

  /**
   * A test to ensure the index is created and items indexed.
   */
  public function testIndex(): void {
    $destination = $this->getDestination();
    /** @var \Drupal\data_pipelines_opensearch\Plugin\DatasetDestination\OpenSearch $destination_plugin */
    $destination_plugin = $destination->getDestinationPlugin();
    // Check the index is empty.
    $this->assertEquals(0, $destination_plugin->getLastDelta($this->dataset));

    // Index the items for the first time.
    $queue_id = $this->dataset->getProcessingQueueId();
    $process_queue = \Drupal::queue($queue_id);
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(1, $destination_plugin->getLastDelta($this->dataset));

    // Add a valid item to the dataset and attempt to index.
    $csv_text = $this->dataset->get('csv_text')->get(0)->getString();
    $csv_text .= <<<CSV
\nY,bob,dylan
CSV;
    $this->dataset->set('csv_text', $csv_text);
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(2, $destination_plugin->getLastDelta($this->dataset));

    // Add an invalid item to the dataset and attempt to index.
    $csv_text = $this->dataset->get('csv_text')->get(0)->getString();
    $invalid_line = "\nY,,knopfler";
    $this->dataset->set('csv_text', $csv_text . $invalid_line);
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(2, $destination_plugin->getLastDelta($this->dataset));

    // Change the invalid item to valid in the dataset and attempt to index.
    $csv_text = $this->dataset->get('csv_text')->get(0)->getString();
    $valid_line = "\nY,mark,knopfler";
    $this->dataset->set('csv_text', str_replace($invalid_line, $valid_line, $csv_text));
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(3, $destination_plugin->getLastDelta($this->dataset));

    // Change the valid item to invalid in the dataset and attempt
    // to index (retain value).
    $csv_text = $this->dataset->get('csv_text')->get(0)->getString();
    $this->dataset->set('csv_text', str_replace($valid_line, $invalid_line, $csv_text));
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(3, $destination_plugin->getLastDelta($this->dataset));

    // Attempt the same action, but with the invalid values handling set
    // to 'remove' (remove value).
    $this->dataset->set('invalid_values', DatasetInterface::INVALID_REMOVE);
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::ProcessCleanup,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(2, $destination_plugin->getLastDelta($this->dataset));
  }

  /**
   * A test to ensure the index can be deleted.
   */
  public function testDelete(): void {
    $destination = $this->getDestination();
    /** @var \Drupal\data_pipelines_opensearch\Plugin\DatasetDestination\OpenSearch $destination_plugin */
    $destination_plugin = $destination->getDestinationPlugin();

    // Index some items.
    $queue_id = $this->dataset->getProcessingQueueId();
    $process_queue = \Drupal::queue($queue_id);
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(1, $destination_plugin->getLastDelta($this->dataset));

    // Perform the deletion. Returning 0 indicates either the index
    // is empty or doesn't exist.
    $this->dataset->delete();
    $this->assertEquals(0, $destination_plugin->getLastDelta($this->dataset));
  }

  /**
   * A test to ensure the index is created with the destination settings.
   */
  public function testDestinationSettings(): void {
    $destination = $this->getDestination();
    /** @var \Drupal\data_pipelines_opensearch\Plugin\DatasetDestination\OpenSearch $destination_plugin */
    $destination_plugin = $destination->getDestinationPlugin();

    // Index some items.
    $queue_id = $this->dataset->getProcessingQueueId();
    $process_queue = \Drupal::queue($queue_id);
    $this->dataset->queueProcessing(
      $process_queue,
      new DatasetBatchContext($this->dataset->getBatchSize())
    );
    $this->processQueueWithWait(
      $queue_id,
      $destination->id(),
      [
        ProcessingOperationEnum::Begin,
        ProcessingOperationEnum::ProcessChunk,
        ProcessingOperationEnum::End,
      ]
    );
    $this->assertEquals(1, $destination_plugin->getLastDelta($this->dataset));

    // Create our own client to query the field mapping.
    /** @var \Drupal\data_pipelines_opensearch\ClientFactoryInterface $client_factory */
    $client_factory = $this->container->get('data_pipelines_opensearch.client_factory');
    $client = $client_factory->create(
      $destination_plugin->getUrl(),
      $destination_plugin->getUsername(),
      $destination_plugin->getPassword(),
    );
    $field_mapping = $client->indices()->getFieldMapping([
      'index' => $destination_plugin->getIndexName($this->dataset),
      'fields' => ['firstname'],
    ]);
    $this->assertEquals(
      'keyword',
      $field_mapping[$destination_plugin->getIndexName($this->dataset)]['mappings']['firstname']['mapping']['firstname']['type']
    );
  }

  /**
   * A test to ensure the last delta returns 0 when there is no mapping info.
   *
   * This would only be an occurrence when an index exists without any items.
   */
  public function testNoLastDelta(): void {
    $destination = $this->getDestination();
    /** @var \Drupal\data_pipelines_opensearch\Plugin\DatasetDestination\OpenSearch $destination_plugin */
    $destination_plugin = $destination->getDestinationPlugin();
    // Create our own client to create the index.
    /** @var \Drupal\data_pipelines_opensearch\ClientFactoryInterface $client_factory */
    $client_factory = $this->container->get('data_pipelines_opensearch.client_factory');
    $client = $client_factory->create(
      $destination_plugin->getUrl(),
      $destination_plugin->getUsername(),
      $destination_plugin->getPassword(),
    );

    // Create an empty index.
    $index_request = new IndexRequest($client);
    $index = new Index($destination_plugin->getIndexName($this->dataset));
    $index_request->create($index);

    // Attempt to get the last delta.
    $last = $destination_plugin->getLastDelta($this->dataset);
    $this->assertEquals(0, $last);
  }

}
